#' @description init the detail table
#' @title  initial the detail frame 
#' @param init_node the start node to find the shortest path, input should be one scalar of the graph
#' @param graph the structure of the graph, input should be data frame or matrix
#' eg:wiki_graph <- data.frame(v1=c(1,1,1,2,2,2,3,3,3,3,4,4,4,5,5,6,6,6),
#'            v2=c(2,3,6,1,3,4,1,2,4,6,2,3,5,4,6,1,3,5),
#'             w=c(7,9,14,7,10,15,9,10,11,2,15,11,6,6,9,14,2,9))
init_node_distance_list <- function (init_node, graph) {
  node_distance_list <- list(init_node = 0)
  # first, get all unique nodes 
  all_nodes <- unique(append(graph[,'v1'],graph[,'v2']))
  if (!init_node %in% all_nodes) {
    stop('node is not in graph')
  }
  all_weights <- unique(graph[,'w'])
  if (0 %in% all_weights) {
    stop('the weight only exist bwtween two different nodes and should not be zero')
  }
  dup = duplicated(graph[c("v1", "v2")])
  node_dup_len = dup[dup==TRUE]
  
  #built list and data frame for later comparison
  detail_frame <- data.frame(matrix(ncol=length(all_nodes),nrow = 3))
  names(detail_frame) <- all_nodes
  row.names(detail_frame) <- c('is_best_node','node_distance','previous_node')
  detail_frame['node_distance', ]<- Inf
  detail_frame['node_distance',init_node] <- 0
  detail_frame['is_best_node',init_node] <- TRUE
  checked_nodes <- list(init_node)
  unchecked_nodes <- all_nodes[-which( all_nodes == init_node)]
  return (list(detail_frame=detail_frame, checked_nodes=checked_nodes, unchecked_nodes=unchecked_nodes))
}


#' @description
#' the update method for detail table to record the shortest path
#' @title update the detail frame when traversing to a new node
#' @param node the current node when traversing all the nodes to find the best way
#' @param init_node the start node in the algorithm
#' @param checked_nodes the nodes that having been tested to find the shortest path
#' @param unchecked_nodes the nodes that having been tested to find the shortest path
#' @param detail_frame to store the shortest path , if it is the best node and the related distance
#' @param graph the graph indicates the relationship of different nodes

update_detail_path <- function(node,init_node,detail_frame,graph,checked_nodes,unchecked_nodes) {
  
  #get connected nodes to node
  if (length(unchecked_nodes) == 0){
    # browser() 
    return (detail_frame)
  }
  connected_nodes <- graph[graph$v1 == node,]$v2
  next_node <- NULL
  min_path <- Inf
  for (c_node in connected_nodes){
    current_distance <- graph[graph$v1 == node & graph$v2 == c_node,]$w 
    if (node == init_node) {
      detail_frame['node_distance',c_node] <- current_distance
      detail_frame['previous_node',c_node] <- init_node
      if(current_distance < min_path) {
        min_path <- current_distance
        next_node <- c_node
        # detail_frame['best_node',c_node] <-TRUE
      }
    }else{
      # browser() 
      if (c_node %in% checked_nodes) next
      
      previous_distance <- detail_frame['node_distance',node] 
      source_distance <-  previous_distance + current_distance
      exist_distance <- detail_frame['node_distance',c_node]
      #差更新前置节点
      if (source_distance < exist_distance) {
        
        # update source distance
        detail_frame['node_distance',c_node] <- source_distance
        # update previous node
        detail_frame['previous_node',c_node] <- node
      }else{
        source_distance <- exist_distance
      }
      if (source_distance < min_path) {
        min_path <- source_distance
        next_node <- c_node
      }
    }
  }
  
  checked_nodes <- append(checked_nodes,next_node)
  unchecked_nodes <- unchecked_nodes[-which(unchecked_nodes==next_node)]
  detail_frame['is_best_node',next_node] <-TRUE
  return (update_detail_path(next_node,init_node,detail_frame,graph,checked_nodes,unchecked_nodes))
}

#' @description
#' the dijkstra algorithm to find the shortest path in the graph
#' @title the entrance of the dijkstra algorithm
#' @param graph a graph to find the shortest path 
#' @param init_node the start node to find the shortest path in the graph 
#' @export
#' @examples
#' # Example 1: Define a graph and use Dijkstra algorithm to find the shortest path
#' wiki_graph <- data.frame(v1=c(1,1,1,2,2,2,3,3,3,3,4,4,4,5,5,6,6,6),
#'                          v2=c(2,3,6,1,3,4,1,2,4,6,2,3,5,4,6,1,3,5),
#'                          w=c(7,9,14,7,10,15,9,10,11,2,15,11,6,6,9,14,2,9))
#' dijkstra(wiki_graph, 3)
dijkstra <- function(graph, init_node) {
  
  res<- init_node_distance_list(init_node,graph)
  detail_frame <- res$detail_frame
  checked_nodes <- res$checked_nodes
  unchecked_nodes <- res$unchecked_nodes
  updated_detail_frame <- update_detail_path(init_node,init_node,detail_frame,graph,checked_nodes,unchecked_nodes)
  return(unname(unlist(updated_detail_frame['node_distance',])))
}
